from kedro_databricks.cli.bundle import bundle
from kedro_databricks.cli.deploy import deploy
from kedro_databricks.cli.destroy import destroy
from kedro_databricks.cli.init import init
from kedro_databricks.cli.run import run

__all__ = ["init", "bundle", "deploy", "run", "destroy"]



from __future__ import annotations

import shutil

from kedro.framework.startup import ProjectMetadata
from kedro.pipeline import Pipeline, node

from kedro_databricks.utils import require_databricks_run_script


def reset_init(metadata: ProjectMetadata):
    (metadata.project_path / "databricks.yml").unlink(missing_ok=True)
    shutil.rmtree(metadata.project_path / "conf" / "dev", ignore_errors=True)
    shutil.rmtree(metadata.project_path / "conf" / "prod", ignore_errors=True)


def identity(arg):
    return arg


pipeline = Pipeline(
    [
        node(
            identity,
            ["input"],
            ["intermediate"],
            name="node0",
            tags=["tag0", "tag1"],
        ),
        node(identity, ["intermediate"], ["output"], name="node1"),
        node(identity, ["intermediate"], ["output2"], name="node2", tags=["tag0"]),
        node(
            identity,
            ["intermediate"],
            ["output3"],
            name="node3",
            tags=["tag1", "tag2"],
        ),
        node(identity, ["intermediate"], ["output4"], name="node4", tags=["tag2"]),
    ]
)


def _generate_task(
    task_key: int | str,
    depends_on: list[str] = [],
    conf: str = "conf",
    env: str = "fake_env",
    runtime_params: str = "",
):
    entry_point = "fake-project"
    params = [
        "--nodes",
        task_key,
        "--conf-source",
        "/Workspace/${workspace.file_path}/" + conf,
        "--env",
        "${var.environment}",
    ]

    if require_databricks_run_script():
        entry_point = "databricks_run"
        params = params + ["--package-name", "fake_project"]

    if runtime_params:
        params = params + ["--params", runtime_params]

    task = {
        "task_key": task_key,
        "libraries": [
            {"whl": "../dist/*.whl"},
        ],
        "python_wheel_task": {
            "package_name": "fake_project",
            "entry_point": entry_point,
            "parameters": params,
        },
    }

    if len(depends_on) > 0:
        task["depends_on"] = [{"task_key": dep} for dep in depends_on]

    return task


def generate_workflow(conf="conf"):
    tasks = []

    for i in range(5):
        if i == 0:
            depends_on = []
        else:
            depends_on = ["node0"]
        tasks.append(_generate_task(f"node{i}", depends_on, conf))

    return {
        "name": "workflow1",
        "tasks": tasks,
    }


WORKFLOW = generate_workflow()



from __future__ import annotations

from collections.abc import Iterable, MutableMapping
from typing import Any

from kedro.framework.project import pipelines
from kedro.framework.startup import ProjectMetadata
from kedro.pipeline import Pipeline
from kedro.pipeline.node import Node

from kedro_databricks.cli.bundle.utils import get_entry_point, remove_nulls, sort_dict
from kedro_databricks.constants import TASK_KEY_ORDER, WORKFLOW_KEY_ORDER
from kedro_databricks.logger import get_logger
from kedro_databricks.utils import make_workflow_name, require_databricks_run_script

log = get_logger("bundle").getChild(__name__)


class ResourceGenerator:
    """Generates Databricks resources for the given pipelines.

    Finds all pipelines in the project and generates Databricks asset bundle resources
    for each according to the Databricks REST API
    """

    def __init__(
        self,
        metadata: ProjectMetadata,
        env: str,
        conf_source: str = "conf",
        params: str | None = None,
    ) -> None:
        self.metadata = metadata
        self.env = env
        self.pipelines: MutableMapping = pipelines
        self.remote_conf_dir = f"/Workspace/${{workspace.file_path}}/{conf_source}"
        self.params = params

    def generate_resources(
        self, pipeline_name: str | None = None
    ) -> dict[str, dict[str, Any]]:
        """Generate Databricks resources for the given pipelines.

        Finds all pipelines in the project and generates Databricks asset bundle resources
        for each according to the Databricks REST API

        Args:
            pipeline_name (str | None): The name of the pipeline for which Databricks asset bundle resources should be generated.
                If None, generates all pipelines.

        Returns:
            dict[str, dict[str, Any]]: A dictionary of pipeline names and their Databricks resources
        """
        workflows = {}
        pipeline = self.pipelines.get(pipeline_name)
        if pipeline_name and pipeline:
            log.info(f"Generating resources for pipeline '{pipeline_name}'")
            name = make_workflow_name(self.metadata.package_name, pipeline_name)
            workflows[name] = self._create_workflow(name=name, pipeline=pipeline)
            return self._workflows_to_resources(workflows)
        if pipeline_name:
            raise KeyError(
                f"Pipeline '{pipeline_name}' not found. Available pipelines: {list(self.pipelines.keys())}"
            )

        for pipe_name, pipeline in self.pipelines.items():
            if len(pipeline.nodes) == 0:
                continue
            name = make_workflow_name(self.metadata.package_name, pipe_name)
            workflow = self._create_workflow(name=name, pipeline=pipeline)
            log.debug(f"Workflow '{name}' successfully created.")
            log.debug(workflow)
            workflows[name] = workflow

        return self._workflows_to_resources(workflows)

    def _workflows_to_resources(
        self, workflows: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        """Convert Databricks workflows to Databricks resources.

        Args:
            workflows (Dict[str, Dict[str, Any]]): A dictionary of Databricks workflows

        Returns:
            Dict[str, Dict[str, Any]]: A dictionary of Databricks resources
        """
        resources = {
            name: {"resources": {"jobs": {name: wf}}} for name, wf in workflows.items()
        }
        log.debug(resources)
        log.info("Databricks resources successfully generated.")
        return resources

    def _create_workflow(self, name: str, pipeline: Pipeline) -> dict[str, Any]:
        """Create a Databricks workflow for a given pipeline.

        Args:
            name (str): name of the pipeline
            pipeline (Pipeline): Kedro pipeline object
            env (str): name of the env to be used by the tasks of the workflow

        Returns:
            Dict[str, Any]: a Databricks workflow
        """
        ## Follows the Databricks REST API schema
        ## https://docs.databricks.com/api/workspace/jobs/create
        workflow = {
            "name": name,
            "tasks": [
                self._create_task(node.name, depends_on=deps)
                for node, deps in sorted(pipeline.node_dependencies.items())
            ],
        }
        non_null = remove_nulls(sort_dict(workflow, WORKFLOW_KEY_ORDER))
        if not isinstance(non_null, dict):  # pragma: no cover - this is a type check
            raise RuntimeError("Expected a dict")
        return non_null

    def _create_task(
        self,
        name: str,
        depends_on: Iterable[Node],
    ) -> dict[str, Any]:
        """Create a Databricks task for a given node.

        Args:
            name (str): name of the node
            depends_on (List[Node]): list of nodes that the task depends on
            env (str): name of the env to be used by the task

        Returns:
            Dict[str, Any]: a Databricks task
        """
        ## Follows the Databricks REST API schema. See "tasks" in the link below
        ## https://docs.databricks.com/api/workspace/jobs/create
        entry_point = get_entry_point(self.metadata.project_name)
        params = [
            "--nodes",
            name,
            "--conf-source",
            self.remote_conf_dir,
            "--env",
            "${var.environment}",
        ]

        if require_databricks_run_script():  # pragma: no cover
            entry_point = "databricks_run"
            params = params + ["--package-name", self.metadata.package_name]

        if self.params:
            params = params + ["--params", self.params]

        task = {
            "task_key": name.replace(".", "_"),
            "libraries": [{"whl": "../dist/*.whl"}],
            "depends_on": [
                {"task_key": dep.name.replace(".", "_")}
                for dep in sorted(depends_on, key=lambda dep: dep.name)
            ],
            "python_wheel_task": {
                "package_name": self.metadata.package_name,
                "entry_point": entry_point,
                "parameters": params,
            },
        }

        return sort_dict(task, TASK_KEY_ORDER)



from kedro.framework.startup import ProjectMetadata

from kedro_databricks.logger import get_logger
from kedro_databricks.utils import Command, assert_databricks_cli

log = get_logger("destroy")


def destroy(metadata: ProjectMetadata, env: str, *databricks_args: str):
    """Destroy the Databricks Asset Bundle.

    This function destroys the Databricks Asset Bundle in the current project
    directory. It removes the Databricks configuration file and the Databricks target
    configuration file, and deletes the corresponding resources from DBFS.

    Args:
        metadata (ProjectMetadata): The project metadata.
        env (str): The environment to destroy.
        *databricks_args: Additional arguments to be passed to the `databricks` CLI.

    Raises:
        RuntimeError: If the `databricks` CLI is not installed or the wrong version is used.
    """
    assert_databricks_cli()
    destroy_cmd = ["databricks", "bundle", "destroy"] + list(databricks_args)
    Command(destroy_cmd, log=log, warn=True).run(cwd=metadata.project_path)
    Command(
        [
            "databricks",
            "fs",
            "rm",
            "-r",
            f"dbfs:/FileStore/{metadata.package_name}/{env}",
        ],
        log=log,
        warn=True,
    ).run()



import json
import re
from pathlib import Path

import yaml
from databricks.sdk import WorkspaceClient
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.constants import DEFAULT_TARGET
from kedro_databricks.logger import get_logger
from kedro_databricks.utils import (
    Command,
    get_bundle_name,
    get_targets,
    read_databricks_config,
)

log = get_logger("init")


class DatabricksTarget:
    """Represents a Databricks target for a Kedro project.

    This class is used to create a target configuration for a Databricks Asset Bundle.
    It retrieves metadata about the target from the Databricks CLI and stores
    relevant information such as the bundle name, target name, mode, host, and file path.

    Attributes:
        bundle (str): The name of the Databricks bundle.
        name (str): The name of the target.
        mode (str): The mode of the target (e.g., "development").
        host (str): The host of the Databricks workspace.
        file_path (str): The file path in the Databricks workspace.

    Args:
        bundle (str): The name of the Databricks bundle.
        name (str): The name of the target.
        conf (dict): The configuration dictionary for the target, which may include
            mode and workspace information.

    Raises:
        ValueError: If the metadata for the target cannot be retrieved.
    """

    def __init__(self, bundle: str, name: str, conf: dict):
        self.bundle = bundle
        self.name = name
        self.mode = conf.get("mode", "development")
        workspace_conf = conf.get("workspace", {})
        self.host = workspace_conf.get("host")
        metadata = self._get_metadata()
        self.file_path = metadata.get("workspace", {}).get("file_path")

    def _get_metadata(self):
        result = Command(
            [
                "databricks",
                "bundle",
                "validate",
                "--target",
                self.name,
                "--output",
                "json",
            ],
            log=log,
            warn=True,
        ).run()
        json_start = [
            i for i in range(len(result.stdout)) if result.stdout[i].startswith("{")
        ]
        if not json_start:  # pragma: no cover
            raise ValueError(f"Could not get metadata for target {self.name}")
        json_output = "\n".join(result.stdout[json_start[0] :])
        return json.loads(json_output)


def create_target_configs(
    metadata: ProjectMetadata,
    node_type_id: str,
    default_key: str,
    single_user_default: bool = True,
):
    """Create target configurations for a Kedro project in Databricks.

    This function creates target configurations for each target defined in the
    Databricks configuration file. It sets up the necessary directories and files
    for each target, including a `.gitkeep` file to ensure the directory is tracked
    by Git. It also creates a target configuration file with the specified node type
    and default key. If the target is the default target, it sets up a specific file path
    for it in the Databricks File System (DBFS).


    Args:
        metadata (ProjectMetadata): The project metadata containing the project path.
        node_type_id (str): The node type ID for the target configuration.
        default_key (str): The default key to use for the target configuration.
        single_user_default (bool, optional): Whether to set the target as single user by default.
            Defaults to True.

    Raises:
        FileNotFoundError: If the Databricks configuration file does not exist.
        ValueError: If the Databricks configuration is invalid or missing required fields.
    """
    conf_dir = metadata.project_path / "conf"
    databricks_config = read_databricks_config(metadata.project_path)
    bundle_name = get_bundle_name(databricks_config)
    targets = get_targets(databricks_config)
    for target_name, target_conf in targets.items():
        target = DatabricksTarget(bundle_name, target_name, target_conf)
        target_conf_dir = conf_dir / target.name
        target_conf_dir.mkdir(exist_ok=True)
        _save_gitkeep_file(target_conf_dir)
        is_single_user = single_user_default and target.name == DEFAULT_TARGET
        target_config = _create_target_config(
            default_key,
            node_type_id,
            single_user=is_single_user,
        )
        _save_target_config(target_config, target_conf_dir)
        target_file_path = f"/Volumes/<your-volume-name>/{bundle_name}/{target_name}"
        if target.name == DEFAULT_TARGET:
            target_file_path = f"/dbfs/FileStore/{bundle_name}/{target_name}"
        _save_target_catalog(conf_dir, target_conf_dir, target_file_path)
        log.info(f"Created target config for {target.name} at {target_conf_dir}")


def _create_target_config(
    default_key: str, node_type_id: str, single_user: bool = False
):
    new_cluster = {
        "spark_version": "15.4.x-scala2.12",
        "node_type_id": node_type_id,
        "num_workers": 1,
        "spark_env_vars": {
            "KEDRO_LOGGING_CONFIG": "/Workspace/\\${workspace.file_path}/conf/logging.yml"
        },
    }

    if single_user:
        wc = WorkspaceClient()
        single_user_opts = {
            "data_security_mode": "SINGLE_USER",
            "single_user_name": wc.current_user.me().user_name,
        }
        new_cluster.update(single_user_opts)

    return {
        default_key: {
            "job_clusters": [
                {"job_cluster_key": default_key, "new_cluster": new_cluster}
            ],
            "tasks": [{"task_key": default_key, "job_cluster_key": default_key}],
        }
    }


def _substitute_file_path(string: str) -> str:
    """Substitute the file path in the catalog"""
    match = re.sub(
        r"(.*:)(.*)(data/.*)",
        r"\g<1> ${_file_path}/\g<3>",
        string,
    )
    return match


def _save_target_catalog(
    conf_dir: Path, target_conf_dir: Path, target_file_path: str
):  # pragma: no cover
    with open(f"{conf_dir}/base/catalog.yml") as f:
        cat = f.read()
    target_catalog = _substitute_file_path(cat)
    with open(target_conf_dir / "catalog.yml", "w") as f:
        f.write("_file_path: " + target_file_path + "\n" + target_catalog)


def _save_target_config(target_config: dict, target_conf_dir: Path):  # pragma: no cover
    with open(target_conf_dir / "databricks.yml", "w") as f:
        yaml.dump(target_config, f)


def _save_gitkeep_file(target_conf_dir: Path):
    if not (target_conf_dir / ".gitkeep").exists():
        with open(target_conf_dir / ".gitkeep", "w") as f:
            f.write("")



from pathlib import Path

import mkdocs_gen_files

mkdocs_gen_files.log.info("Generating changelog...")
root = Path(__file__).parent.parent

with mkdocs_gen_files.open("changelog.md", "w") as f:
    changelog_file = root / "CHANGELOG.md"
    f.write(changelog_file.read_text())



import pytest

from kedro_databricks.constants import NODE_TYPE_MAP
from kedro_databricks.plugin import commands
from kedro_databricks.utils import get_bundle_name, get_targets, read_databricks_config
from tests.utils import reset_init


@pytest.mark.parametrize("provider", ["azure", "aws", "gcp", "unknown"])
def test_init_arg(cli_runner, metadata, provider):
    reset_init(metadata)
    command = ["databricks", "init", "--provider", provider]
    result = cli_runner.invoke(commands, command, obj=metadata)
    if provider not in NODE_TYPE_MAP:
        assert result.exit_code == 1, (result.exit_code, result.stdout)
    else:
        assert result.exit_code == 0, (result.exit_code, result.stdout)
        files = [f"{f.parent.name}/{f.name}" for f in metadata.project_path.rglob("*")]
        assert len(files) > 0, "Found no files in the directory."

        config_path = metadata.project_path / "databricks.yml"
        assert config_path.exists(), f"Configuration at {config_path} does not exist"

        databricks_config = read_databricks_config(metadata.project_path)
        assert databricks_config is not None, "Databricks config not read"
        name = get_bundle_name(databricks_config)
        assert name == metadata.package_name, f"Bundle name not set: {name}"

        targets = get_targets(databricks_config)
        for target in targets:
            override_path = metadata.project_path / "conf" / target / "databricks.yml"
            assert (
                override_path.exists()
            ), f"Resource Overrides at {override_path} does not exist"



from kedro.framework.startup import ProjectMetadata

from kedro_databricks.logger import get_logger
from kedro_databricks.utils import Command, assert_databricks_cli

log = get_logger("run")


def run(metadata: ProjectMetadata, pipeline: str, *databricks_args: str):
    """Run a Databricks job using the specified pipeline.

    This function runs a Databricks job using the specified pipeline and additional
    arguments. It assumes that the Databricks CLI is installed and configured correctly.

    Args:
        metadata (ProjectMetadata): The project metadata.
        pipeline (str): The name of the pipeline to run.
        *databricks_args: Additional arguments to be passed to the `databricks` CLI.

    Raises:
        RuntimeError: If the `databricks` CLI is not installed or the job fails to run.
    """
    assert_databricks_cli()
    cmd = ["databricks", "bundle", "run", pipeline] + list(databricks_args)
    log.info(f"Running `{' '.join(cmd)}` in {metadata.project_path}")
    result = Command(cmd, log=log, warn=True).run(cwd=metadata.project_path)
    if result.returncode != 0:  # pragma: no cover
        raise RuntimeError("Failed to run Databricks job\n" + "\n".join(result.stdout))



from __future__ import annotations

import logging
import platform
import subprocess

import pytest

from kedro_databricks.constants import MINIMUM_DATABRICKS_VERSION
from kedro_databricks.utils import (
    Command,
    _version_to_str,
    assert_databricks_cli,
    get_bundle_name,
    get_targets,
    make_workflow_name,
    read_databricks_config,
)

log = logging.getLogger("test")


@pytest.mark.skipif(
    assert_databricks_cli(False) is not None,
    reason="Databricks CLI is not installed",
)
def test_assert_databricks_cli():
    assert_databricks_cli()


def test_read_databricks_config(metadata):
    if not (metadata.project_path / "databricks.yml").exists():
        (metadata.project_path / "databricks.yml").write_text(
            "bundle:\n  name: test_bundle\n"
        )
    read_databricks_config(metadata.project_path)


def test_read_databricks_config_does_not_exist(metadata):
    if (metadata.project_path / "databricks.yml").exists():
        (metadata.project_path / "databricks.yml").unlink()
    with pytest.raises(FileNotFoundError):
        read_databricks_config(metadata.project_path)


def test_read_bundle_name_does_not_exist():
    with pytest.raises(ValueError):
        get_bundle_name({})


def test_read_bundle_name():
    name = get_bundle_name({"bundle": {"name": "test_bundle"}})
    assert name == "test_bundle", name


def test_get_targets():
    targets = get_targets({"targets": {"test": "test"}})
    assert targets == {"test": "test"}, targets


def test_get_targets_does_not_exist():
    with pytest.raises(ValueError):
        get_targets({})


@pytest.mark.parametrize(
    ["version", "expected", "raises"],
    [
        ([1, 2, 3], "1.2.3", False),
        ([1, 2], "1.2", True),
        ([1], "1", True),
        ([1, 2, 3, 4], "1.2.3.4", True),
    ],
)
def test_version_to_str(version, expected, raises):
    if raises is True:
        with pytest.raises(ValueError):
            _version_to_str(version)
    else:
        result = _version_to_str(version)
        assert result == expected, f"Expected {expected}, but got {result}"


def test_fail_with_python_databricks_cli(monkeypatch):
    subprocess.run(["uv", "pip", "install", "databricks-cli"], check=True)
    monkeypatch.setenv("DATABRICKS_CLI_DO_NOT_EXECUTE_NEWER_VERSION", "1")
    with pytest.raises(
        RuntimeError,
        match=f"this script requires at least {'.'.join(map(str, MINIMUM_DATABRICKS_VERSION))}",
    ):
        assert_databricks_cli()
    subprocess.run(["uv", "pip", "uninstall", "databricks-cli"], check=True)


def test_fail_with_python_databricks_cli_soft(monkeypatch):
    subprocess.run(["uv", "pip", "install", "databricks-cli"], check=True)
    monkeypatch.setenv("DATABRICKS_CLI_DO_NOT_EXECUTE_NEWER_VERSION", "1")
    err = assert_databricks_cli(False)
    assert err is not None
    subprocess.run(["uv", "pip", "uninstall", "databricks-cli"], check=True)


OS = platform.uname().system.lower()


@pytest.mark.parametrize(
    ["cmd", "result_code", "warn", "raises"],
    [
        (["ls", "."], 0, False, False),
        (["ls", "non_existent_file"], 2, False, True),
        (["ls", "non_existent_file"], 1 if OS == "darwin" else 2, True, False),
        (["ls", "non_existent_file"], 2, False, True),
        (["ls", "non_existent_file"], 1 if OS == "darwin" else 2, True, False),
    ],
)
def test_command(cmd, result_code, warn, raises):
    if raises and not warn:
        with pytest.raises(Exception):
            result = Command(cmd, log=log).run()
            assert isinstance(result, subprocess.CompletedProcess)
            assert result.returncode == result_code, result
    else:
        command = Command(cmd, log=log, warn=warn)
        assert repr(command) == f"Command({cmd})", repr(command)
        result = command.run()
        assert isinstance(result, subprocess.CompletedProcess)
        assert result.returncode == result_code, result


@pytest.mark.parametrize(
    ["package_name", "pipeline_name", "expected"],
    [
        ("package", "__default__", "package"),
        ("package", "pipeline", "package_pipeline"),
    ],
)
def test_make_workflow_name(package_name, pipeline_name, expected):
    workflow_name = make_workflow_name(package_name, pipeline_name)
    assert workflow_name == expected, workflow_name



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from subprocess import CompletedProcess
from typing import cast

import pytest
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.cli.deploy import (
    _build_project,
    _check_deployment_complete,
    _get_arg_value,
    _validate_project,
)
from kedro_databricks.cli.deploy.get_deployed_resources import (
    _get_username,
    _is_valid_job,
)


@dataclass
class MockResult:
    stdout: list[str]
    stderr: list[str]


class MetadataMock:
    def __init__(self, path: str, name: str):
        self.project_path = Path(path)
        self.project_name = name
        self.package_name = name
        self.source_dir = "src"
        self.env = "local"
        self.config_file = "conf/base"
        self.project_version = "0.16.0"
        self.project_description = "Test Project Description"
        self.project_author = "Test Author"
        self.project_author_email = "author@email.com"


def test_deploy_validate_project(metadata):
    with pytest.raises(
        FileNotFoundError, match="Databricks configuration file does not exist."
    ):
        _validate_project(metadata)


def test_deploy_non_existing_path(metadata):
    with pytest.raises(FileNotFoundError, match="Project path.*does not exist"):
        _validate_project(
            cast(
                ProjectMetadata,  # noqa: F821
                MetadataMock(
                    "/tmp/non_existent_path" + str(os.getpid()), "non_existent_project"
                ),
            )
        )


def test_deploy_valid_project(metadata):
    Path(metadata.project_path / "databricks.yml").write_text("")
    _validate_project(metadata)


def test_deploy_build_project(metadata):
    result = _build_project(metadata)
    assert result.returncode == 0, (result.returncode, result.stdout)


@pytest.mark.parametrize(
    "args, arg, expected",
    [
        (["--env", "local"], "--env", "local"),
        (["--env", "dev"], "--env", "dev"),
        (["--env", "prod"], "--env", "prod"),
        (["--target", "local"], "--target", "local"),
        (["--target", "dev"], "--target", "dev"),
        (["--target", "prod"], "--target", "prod"),
        (["my-program", "--arg1", "value1", "--arg2", "value2"], "--arg1", "value1"),
        (["my-program", "--arg1", "value1", "--arg2", "value2"], "--arg2", "value2"),
        (["my-program", "--arg1=value1"], "--arg1", "value1"),
    ],
)
def test_get_arg_value(args, arg, expected):
    """Test the function to get the value of a specific argument from a list of arguments."""
    result = _get_arg_value(args, arg)
    assert result == expected, f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "result, expected",
    [
        (MockResult([], []), False),
        (MockResult(["Deployment complete!"], []), True),
        (MockResult([], ["Deployment complete!"]), True),
    ],
)
def test_check_deployment_complete(result, expected):
    assert _check_deployment_complete(cast(CompletedProcess, result)) == expected


@pytest.mark.parametrize(
    "pipelines, job_name, expected",
    [
        ({"__default__": {}}, "fake_project", True),
        ({"test_job": {}}, "fake_project_test_job", True),
        ({"test_job": {}}, "test_job", False),
    ],
)
def test_is_valid_job(metadata, pipelines, job_name, expected):
    is_valid = _is_valid_job(metadata, pipelines, job_name)
    assert is_valid is expected, "Job should be valid"


@pytest.mark.parametrize(
    "username, raises, expected",
    [
        ("user@email", False, "user"),
        ("test", False, "test"),
        ("test@", False, "test"),
        (None, True, None),
    ],
)
def test_get_username(username, raises, expected):
    username = _get_username("user@email.com")
    assert username == "user", f"Expected 'user', but got {username}"



"""This module provides functionality to bundle Kedro pipelines into Databricks asset bundle resources."""

from __future__ import annotations

from typing import Any

import yaml
from kedro.config import MissingConfigException
from kedro.framework.session import KedroSession
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.cli.bundle.generate_resources import ResourceGenerator
from kedro_databricks.cli.bundle.override_resources import override_resources
from kedro_databricks.logger import get_logger

log = get_logger("bundle")


def bundle(
    metadata: ProjectMetadata,
    env: str,
    default_key: str,
    conf_source: str = "conf",
    params: str | None = None,
    pipeline_name: str | None = None,
    overwrite: bool = False,
):
    """Convert Kedro pipelines into Databricks asset bundle resources.

    This function generates asset bundle resources for the specified Kedro pipeline
    and saves them in the `resources` directory. To validate the generated resources,
    you can run the `databricks bundle validate` command.

    Args:
        metadata (ProjectMetadata): The metadata of the project.
        env (str): The environment for the Kedro project (e.g., "dev", "prod").
        default_key (str): The default configuration key to use for Databricks.
        conf_source (str): The source of the Kedro configuration files (default: "conf").
        params (str | None): Kedro run time parameters in `key1=value1,key2=value2` format (optional).
        pipeline_name (str | None): The pipeline to bundle (optional).
        overwrite (bool): Whether to overwrite existing resources (default: False).
    """
    if default_key.startswith("_"):  # pragma: no cover
        raise ValueError(
            "Default key cannot start with `_` as this is not recognized by OmegaConf."
        )
    overrides = _load_kedro_env_config(metadata, config_dir=conf_source, env=env)
    g = ResourceGenerator(metadata, env, conf_source, params)
    resources = g.generate_resources(pipeline_name)
    result = {}
    for name, resource in resources.items():
        result[name] = override_resources(resource, overrides, default_key)
    _save_bundled_resources(metadata, result, overwrite)


def _load_kedro_env_config(
    metadata: ProjectMetadata, config_dir: str, env: str
) -> dict[str, Any]:
    """Load the Databricks configuration for the given environment.

    Args:
        metadata (ProjectMetadata): The metadata of the project
        config_dir (str): The name of the configuration directory
        env (str): The name of the kedro environment
        MSG (str): The message to display

    Returns:
        dict[str, Any]: The Databricks configuration for the given environment
    """
    local_config_dir = metadata.project_path / config_dir / env

    # If the configuration directory does not exist, Kedro will not load any configuration
    if not local_config_dir.exists():
        log.warning(f"Creating {local_config_dir.relative_to(metadata.project_path)}")
        local_config_dir.mkdir(parents=True)

    with KedroSession.create(project_path=metadata.project_path, env=env) as session:
        config_loader = session._get_config_loader()
        # Backwards compatibility for ConfigLoader that does not support `config_patterns`
        if not hasattr(config_loader, "config_patterns"):
            return config_loader.get("databricks*", "databricks/**")  # pragma: no cover

        # Set the default pattern for `databricks` if not provided in `settings.py`
        if "databricks" not in config_loader.config_patterns.keys():  # type: ignore
            config_loader.config_patterns.update(  # pragma: no cover # type: ignore
                {"databricks": ["databricks*", "databricks/**"]}
            )

        assert "databricks" in config_loader.config_patterns.keys()  # type: ignore

        # Load the config
        try:
            return config_loader["databricks"]
        except MissingConfigException:  # pragma: no cover
            log.warning("No Databricks configuration found.")
            return {}


def _save_bundled_resources(
    metadata: ProjectMetadata,
    resources: dict[str, dict[str, Any]],
    overwrite: bool = False,
):
    """Save the generated resources to the project directory.

    Args:
        resources (Dict[str, Dict[str, Any]]): A dictionary of pipeline names and their Databricks resources
        metadata (ProjectMetadata): The metadata of the project
        overwrite (bool): Whether to overwrite existing resources
    """
    resources_dir = metadata.project_path / "resources"
    resources_dir.mkdir(exist_ok=True, parents=True)
    for name, resource in resources.items():
        p = resources_dir / f"{name}.yml"
        relative_path = p.relative_to(metadata.project_path)

        if p.exists() and not overwrite:  # pragma: no cover
            log.warning(f"{relative_path} already exists. Use --overwrite to replace.")
            continue

        with open(p, "w") as f:
            yaml.dump(resource, f, default_flow_style=False, indent=4, sort_keys=False)
            log.info(f"Wrote {relative_path}")



"""Logger configuration for the kedro-databricks package.

This module sets up a logger for the kedro-databricks package, allowing for structured logging
and easier debugging. The logger is configured to log messages at the INFO level by default,
but this can be overridden by setting the `LOG_LEVEL` environment variable.
"""

import logging
import os

logging.basicConfig(
    level=os.getenv("LOG_LEVEL", "INFO"),
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


def get_logger(name: str) -> logging.Logger:
    """Get a logger with the specified name.

    This function retrieves a logger instance with the given name, which is a child of the root logger
    for the `kedro-databricks` package

    Args:
        name (str): The name of the logger.

    Returns:
        logging.Logger: The logger instance.
    """
    ROOT_LOGGER = logging.getLogger("kedro-databricks")
    return ROOT_LOGGER.getChild(name)



from __future__ import annotations

import copy
import re
from typing import Any


def remove_nulls(value: dict[str, Any] | list[Any]) -> dict[str, Any] | list[Any]:
    """Remove None values from a dictionary or list.

    Args:
        value (Dict[Any, Any] | List[Dict[Any, Any]]): dictionary or list to remove None values from

    Returns:
        Dict[Any, Any] | List[Dict[Any, Any]]: dictionary or list with None values removed
    """
    non_null = copy.deepcopy(value)
    if isinstance(non_null, dict):
        _remove_nulls_from_dict(non_null)
    elif isinstance(non_null, list):
        _remove_nulls_from_list(non_null)
    return non_null


def _remove_nulls_from_list(lst: list) -> list:
    """Remove None values from a list.

    Args:
        l (List[Dict[Any, Any]]): list to remove None values from
    """
    for i, item in enumerate(lst):
        value = remove_nulls(item)
        if not value:
            del lst[i]
        else:
            lst[i] = value
    return lst


def _remove_nulls_from_dict(d: dict) -> dict:
    """Remove None values from a dictionary.

    Args:
        d (Dict[str, Any]): dictionary to remove None values from

    Returns:
        Dict[str, float | int | str | bool]: dictionary with None values removed
    """
    for k, v in list(d.items()):
        value = remove_nulls(v)
        if not value:
            del d[k]
        else:
            d[k] = value
    return d


def get_entry_point(project_name: str) -> str:
    """Get the entry point for a project.

    Args:
        project_name (str): name of the project

    Returns:
        str: entry point for the project
    """
    entrypoint = project_name.strip().lower()
    entrypoint = re.sub(r" +", " ", entrypoint)
    entrypoint = re.sub(r"[^a-zA-Z]", "-", entrypoint)
    entrypoint = re.sub(r"(-+)$", "", entrypoint)
    entrypoint = re.sub(r"^(-+)", "", entrypoint)
    return entrypoint


def sort_dict(d: dict[Any, Any], key_order: list[str]) -> dict[Any, Any]:
    """Recursively sort the keys of a dictionary.

    Args:
        d (Dict[Any, Any]): dictionary to sort
        key_order (List[str]): list of keys to sort by

    Returns:
        Dict[Any, Any]: dictionary with ordered values
    """
    other_keys = [k for k in d.keys() if k not in key_order]
    order = key_order + other_keys

    return dict(sorted(d.items(), key=lambda x: order.index(x[0])))



from pathlib import Path

import mkdocs_gen_files

mkdocs_gen_files.log.info("Generating example pages...")
root = Path(__file__).parent.parent
examples_dir = root / "examples"


def gen():
    """Generate the code reference pages."""
    for p in sorted(examples_dir.iterdir()):
        example_name = p.name
        if not p.is_dir() or example_name.startswith("."):
            continue

        example = _load_example(p)

        with mkdocs_gen_files.open(f"examples/{example_name}.md", "w") as f:
            f.write("\n".join(example) + "\n")
    pass


def _load_example(p: Path) -> list[str]:
    example_name = p.name
    description = ""
    if (p / "README.md").exists():
        description = (p / "README.md").read_text().strip()

    databricks_config = _parse_file(p / "databricks.yml")
    resources_config = _parse_file(p / "resources.yml")
    result_config = _parse_file(p / "result.yml")
    result_diff = _add_diff(resources_config, result_config)

    parts = [
        f"# {example_name.title().replace('_', ' ')}",
        "",
        description,
        "",
        '=== "conf/[env]/databricks.yml"',
        "    ```yaml",
        *_add_padding(databricks_config, leftpad=4),
        "    ```",
        "",
        '=== "Before: resources/<pipeline>.yml"',
        "    ```yaml",
        *_add_padding(resources_config, leftpad=4),
        "    ```",
        "",
        '=== "After: resources/<pipeline>.yml"',
        "    ```diff",
        *_add_padding(result_diff, leftpad=4),
        "    ```",
    ]
    return parts


def _add_diff(old, new):
    i, j = 0, 0
    diffed = []
    while j < len(new):
        i = min(i, len(old) - 1)
        if old[i] != new[j]:
            diffed.append("+" + new[j][1:])
            j += 1
        elif old[i] == new[j]:
            diffed.append(new[j])
            i += 1
            j += 1
        else:
            raise ValueError("Unexpected case in diffing")
    return diffed


def _parse_file(file_path: Path, leftpad: int = 4) -> list[str]:
    """Parse a file and return its content."""
    return [line for line in file_path.read_text().strip().splitlines()]


def _add_padding(lines: list[str], leftpad: int = 4) -> list[str]:
    """Add padding to each line in the list."""
    padding = " " * leftpad
    return [padding + line for line in lines]


gen()



from __future__ import annotations

import logging
import os
import shutil
from pathlib import Path

from dotenv import load_dotenv
from kedro.framework.cli.starters import create_cli as kedro_cli
from kedro.framework.startup import bootstrap_project
from pytest import fixture

from kedro_databricks.constants import DEFAULT_TARGET
from kedro_databricks.plugin import commands

load_dotenv()

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


@fixture(scope="session")
def custom_username():
    custom_username = os.getenv("CUSTOM_USERNAME")
    return custom_username


PROJECT_NAME = "databricks-iris"


@fixture
def databricks_iris_starter(cli_runner):
    project_path = Path().cwd() / PROJECT_NAME
    if project_path.exists():
        shutil.rmtree(project_path)
    cli_runner.invoke(
        # Supply name, tools, and example to skip interactive prompts
        kedro_cli,
        ["new", "-v", "--starter", "databricks-iris", "--name", PROJECT_NAME],
    )
    assert project_path.exists(), "Project path not created"
    assert project_path.is_dir(), "Project path is not a directory"
    # This is needed to ensure that the project is created in the correct directory
    # and not in the current working directory.
    os.chdir(project_path)
    return project_path


@fixture
def iris_meta(databricks_iris_starter):
    # cwd() depends on ^ the isolated filesystem, created by CliRunner()
    project_path = databricks_iris_starter.resolve()
    metadata = bootstrap_project(project_path)
    return metadata


def test_run(cli_runner, iris_meta):
    """Test the run command."""
    project_path = iris_meta.project_path
    provider = "azure"
    command = ["databricks", "init", "--provider", provider]
    result = cli_runner.invoke(commands, command, obj=iris_meta)  # noqa: F821
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert (project_path / "databricks.yml").exists(), "Databricks config not created"
    command = ["databricks", "bundle", "--env", DEFAULT_TARGET]
    result = cli_runner.invoke(commands, command, obj=iris_meta)
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert (project_path / "resources").exists(), "Resources directory not created"
    deploy_cmd = ["databricks", "deploy", "--bundle"]
    result = cli_runner.invoke(commands, deploy_cmd, obj=iris_meta)
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    command = ["databricks", "run", iris_meta.package_name]
    result = cli_runner.invoke(commands, command, obj=iris_meta)
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    command = ["databricks", "destroy", "--auto-approve"]
    result = cli_runner.invoke(commands, command, obj=iris_meta)
    assert result.exit_code == 0, (result.exit_code, result.stdout)



import copy
from typing import Any

from kedro_databricks.constants import IGNORED_OVERRIDE_KEYS, OVERRIDE_KEY_MAP
from kedro_databricks.logger import get_logger

log = get_logger("bundle").getChild(__name__)


def override_resources(bundle: dict, overrides: dict, default_key) -> dict[str, Any]:
    """Override the resources in a Databricks bundle.

    This function applies the given overrides to the resources in a Databricks bundle.

    Args:
        bundle (Dict): the Databricks bundle
        overrides (Dict): the overrides to apply
        default_key (str): the default key to use for overrides

    Raises:
        ValueError: if the workflow or overrides are not dictionaries
        ValueError: if the key in overrides is not found in OVERRIDE_KEY_MAP

    Returns:
        Dict[str, Any]: the Databricks bundle with the overrides applied
    """
    result = {"resources": {"jobs": {}}}
    for name, workflow in bundle.get("resources", {}).get("jobs", {}).items():
        if not isinstance(workflow, dict):
            raise ValueError(f"workflow must be a dictionary not {type(workflow)}")
        if not isinstance(overrides, dict):
            raise ValueError(f"overrides must be a dictionary not {type(overrides)}")
        workflow_overrides, task_overrides = _get_workflow_overrides(
            overrides, workflow, default_key
        )
        result["resources"]["jobs"][name] = _override_workflow(
            workflow=workflow,
            workflow_overrides=workflow_overrides,
            task_overrides=task_overrides,
            default_key=default_key,
        )
    return result


def _get_lookup_key(key: str):
    lookup = OVERRIDE_KEY_MAP.get(key)
    if lookup is None:
        raise ValueError(f"Key {key} not found in OVERRIDE_KEY_MAP")
    return lookup


def _get_defaults(lst: list, lookup_key: str, default_key):
    """Get the default dictionary from a list of dictionaries.

    Args:
        lst (List[Dict]): list of dictionaries
        lookup_key (str): key to use for looking up dictionaries

    Returns:
        Dict: the default dictionary
    """
    for d in lst:
        if isinstance(d, dict) and d.get(lookup_key, "") == default_key:
            return d
    return {}


def _update_list_by_key(
    old: list[dict[str, Any]],
    new: list[dict[str, Any]],
    lookup_key: str,
    default: dict[str, Any] = {},
    default_key: str = "default",
):
    """Update a list of dictionaries with another list of dictionaries.

    Args:
        old (List[Dict[str, Any]]): list of dictionaries to update
        new (List[Dict[str, Any]]): list of dictionaries to update with
        lookup_key (str): key to use for looking up dictionaries
        default (Dict[str, Any], optional): default dictionary to use for updating

    Returns:
        List[Dict[str, Any]]: updated list of dictionaries
    """
    assert isinstance(
        old, list
    ), f"old must be a list not {type(old)} for key: {lookup_key} - {old}"
    assert isinstance(
        new, list
    ), f"new must be a list not {type(new)} for key: {lookup_key} - {new}"
    assert all(
        lookup_key in o for o in old
    ), f"lookup_key {lookup_key} not found in current: {old}"
    assert all(
        lookup_key in n for n in new
    ), f"lookup_key {lookup_key} not found in updates: {new}"

    old_obj = {curr.pop(lookup_key): curr for curr in copy.deepcopy(old)}
    new_obj = {update.pop(lookup_key): update for update in copy.deepcopy(new)}
    keys = set(old_obj.keys()).union(set(new_obj.keys()))

    for key in keys:
        if default and key == default_key:
            continue
        update = copy.deepcopy(default)
        update.pop(lookup_key, None)
        update.update(new_obj.get(key, {}))
        new = _override_workflow(old_obj.get(key, {}), update, {}, default_key)  # type: ignore
        old_obj[key] = new  # type: ignore

    return sorted(
        [{lookup_key: k, **v} for k, v in old_obj.items()], key=lambda x: x[lookup_key]
    )


def _get_old_value(result: Any, key: Any, value: Any):
    default = None
    if isinstance(value, dict):
        default = {}
    elif isinstance(value, list):
        default = []
    return result.get(key, default)


def _get_workflow_overrides(overrides, workflow, default_key):
    default_overrides = overrides.copy().pop(default_key, {})
    workflow_overrides = overrides.copy().pop(workflow.get("name"), {})
    all_overrides = {**default_overrides, **workflow_overrides}
    default_task = _get_defaults(
        all_overrides.get("tasks", []), "task_key", default_key
    )
    return all_overrides, default_task


def _override_workflow(
    workflow: dict,
    workflow_overrides: dict,
    task_overrides: dict = {},
    default_key: str = "default",
):
    """Override a Databricks workflow with the given overrides.

    Args:
        workflow (Dict): the Databricks workflow
        overrides (Dict): the overrides to apply

    Returns:
        Dict: the Databricks workflow with the overrides applied
    """
    result = {**workflow}

    for key, value in workflow_overrides.items():
        old_value = _get_old_value(result, key, value)
        if isinstance(value, dict) and isinstance(old_value, dict):
            result[key] = _override_workflow(
                workflow=old_value,
                workflow_overrides=value,
                task_overrides={},
                default_key=default_key,
            )
        elif isinstance(value, list) and isinstance(old_value, list):
            if isinstance(value[0], dict) and key not in IGNORED_OVERRIDE_KEYS:
                lookup_key = _get_lookup_key(key)
                if lookup_key:
                    result[key] = _update_list_by_key(
                        old=old_value,
                        new=value,
                        lookup_key=lookup_key,
                        default=task_overrides if key == "tasks" else {},
                        default_key=default_key,
                    )
            elif key == "parameters":
                # Special case for parameters, which can be a list of strings
                log.debug(f"Overriding parameters for key {key} with value {value}")
                result[key] = old_value + value
            else:
                result[key] = value
        else:
            result[key] = workflow_overrides.get(key, old_value)
    return result



from pathlib import Path

import mkdocs_gen_files

root = Path(__file__).parent.parent
src = root / "src"

for path in sorted([f for f in src.rglob("*.py") if f.name != "databricks_run.py"]):
    module_path = path.relative_to(src).with_suffix("")
    module_name = module_path.name
    identifier = ".".join([part for part in module_path.parts if part != "__init__"])
    docname = module_name if module_name != "__init__" else "index"
    mkdocs_gen_files.log.info(
        f"{module_path}: Generating reference page for {identifier} - {docname}"
    )

    doc_path = (module_path.parent / docname).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    if module_path.name == "__main__":
        continue

    if identifier in ["kedro_databricks", "kedro_databricks.cli"]:
        continue

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        print("::: " + identifier, file=fd)

    mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))



from kedro.pipeline import Pipeline, node

from kedro_databricks.cli.deploy.get_deployed_resources import get_deployed_resources
from kedro_databricks.constants import DEFAULT_TARGET
from kedro_databricks.plugin import commands
from tests.utils import reset_init


def identity(arg):
    return arg


pipeline = Pipeline(
    [
        node(
            identity,
            ["input"],
            ["intermediate"],
            name="node0",
            tags=["tag0", "tag1"],
        ),
        node(identity, ["intermediate"], ["output"], name="node1"),
        node(identity, ["intermediate"], ["output2"], name="node2", tags=["tag0"]),
        node(
            identity,
            ["intermediate"],
            ["output3"],
            name="node3",
            tags=["tag1", "tag2"],
        ),
        node(identity, ["intermediate"], ["output4"], name="node4", tags=["tag2"]),
    ],
    tags="pipeline0",
)

pipelines = {
    "__default__": pipeline,
    "ds": pipeline,
}


def test_deploy(cli_runner, metadata, custom_username):
    """Test the `deploy` command"""
    reset_init(metadata)
    deploy_fail = ["databricks", "deploy"]
    result = cli_runner.invoke(commands, deploy_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout)

    init_cmd = ["databricks", "init", "--provider", "azure"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = metadata.project_path / "conf" / DEFAULT_TARGET / "databricks.yml"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert metadata.project_path.exists(), "Project path not created"
    assert metadata.project_path.is_dir(), "Project path is not a directory"
    assert override_path.exists(), "Override file not created"

    deploy_cmd = ["databricks", "deploy", "--bundle"]
    result = cli_runner.invoke(commands, deploy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)

    resources = get_deployed_resources(
        metadata, pipelines, only_dev=True, _custom_username=custom_username
    )
    assert len(resources) > 0, f"There are no resources: {resources}"
    assert all(
        metadata.package_name in p.name for p in resources
    ), f"Package name not in resource: {[p.name for p in resources if metadata.package_name not in p.name]}"
    destroy_cmd = ["databricks", "destroy", "--auto-approve"]
    result = cli_runner.invoke(commands, destroy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)


def test_deploy_prod(cli_runner, metadata, custom_username):
    """Test the `deploy` command"""
    reset_init(metadata)
    deploy_fail = ["databricks", "deploy"]
    result = cli_runner.invoke(commands, deploy_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout)

    init_cmd = ["databricks", "init", "--provider", "azure"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = metadata.project_path / "conf" / "prod" / "databricks.yml"
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert metadata.project_path.exists(), "Project path not created"
    assert metadata.project_path.is_dir(), "Project path is not a directory"
    assert override_path.exists(), "Override file not created"

    deploy_cmd = [
        "databricks",
        "deploy",
        "--env",
        "prod",
        "--bundle",
        "--target",
        "prod",
    ]
    result = cli_runner.invoke(commands, deploy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)

    resources = get_deployed_resources(
        metadata, pipelines, _custom_username=custom_username
    )
    assert len(resources) > 0, f"There are no resources: {resources}"
    assert all(
        metadata.package_name in p.name for p in resources
    ), f"Package name not in resource: {[p.name for p in resources if metadata.package_name not in p.name]}"
    destroy_cmd = ["databricks", "destroy", "--target", "prod", "--auto-approve"]
    result = cli_runner.invoke(commands, destroy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)


def test_deploy_with_conf(cli_runner, metadata):
    """Test the `deploy` command"""
    reset_init(metadata)
    deploy_fail = ["databricks", "deploy"]
    result = cli_runner.invoke(commands, deploy_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout)

    CONF_KEY = "custom_conf"

    init_cmd = ["databricks", "init", "--provider", "azure"]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    override_path = metadata.project_path / CONF_KEY / DEFAULT_TARGET / "databricks.yml"
    override_path.parent.mkdir(parents=True, exist_ok=True)
    override_path.write_text(
        """
        default:
            job_clusters:
                - job_cluster_key: default
                new_cluster:
                    spark_version: 14.3.x-scala2.12
                    node_type_id: Standard_DS3_v2
                    num_workers: 1
                    spark_env_vars:
                        KEDRO_LOGGING_CONFIG: "/dbfs/FileStore/develop_eggs/conf/logging.yml"
            tasks:
                - task_key: default
                job_cluster_key: default
        """
    )
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert metadata.project_path.exists(), "Project path not created"
    assert metadata.project_path.is_dir(), "Project path is not a directory"

    settings = metadata.project_path / "src" / metadata.package_name / "settings.py"
    original_settings = settings.read_text()
    with open(settings, "a"):
        settings.write_text(f"CONF_SOURCE = '{CONF_KEY}'")

    deploy_cmd = ["databricks", "deploy", "--bundle", f"--conf-source={CONF_KEY}"]
    result = cli_runner.invoke(commands, deploy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    settings.write_text(original_settings)
    destroy_cmd = ["databricks", "destroy", "--auto-approve"]
    result = cli_runner.invoke(commands, destroy_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout)
